<?php


/**
 * @file
 * Handler class for drupal_database_table storage plugin.
 */

class SamplerStorageHandlerDrupalDatabaseTable implements SamplerStorageHandlerInterface {

  public function __construct($sampler) {
    $this->sampler = $sampler;

    // Dump in plugin option defaults.
    $this->sampler->options = $this->sampler->options + $this->options();
  }

  public function options() {
    return array();
  }

  /**
   * The Drupal table name for the metric.
   *
   * @return
   *   The Drupal table name.
   */
  public function schemaIdentifierDrupal() {
    return $this->schemaIdentifier();
  }

  public function schemaIdentifier() {
    $number_of_values = count($this->sampler->dataType);
    return "sampler_metrics_values_$number_of_values";
  }

  public function reportSchemaToDrupal() {
    return TRUE;
  }

  /**
   * Wrapper functions around Drupal's database layer functions.  These allow
   * other classes to use Drupal's external database functionality via the
   * setDatabaseConnection() method.
   */

  public function tableExists($table) {
    $previous_db = $this->setDatabaseConnection();
    $table_exists = db_table_exists($table);
    $this->setDatabaseConnection($previous_db);
    return $table_exists;
  }

  public function createTable($table, $schema) {
    $ret = array();
    $previous_db = $this->setDatabaseConnection();
    db_create_table($ret, $table, $schema);
    $this->setDatabaseConnection($previous_db);
  }

  public function dropTable($table) {
    $ret = array();
    $previous_db = $this->setDatabaseConnection();
    db_drop_table($ret, $table);
    $this->setDatabaseConnection($previous_db);
  }

  public function databaseQuery() {
    $args = func_get_args();
    $previous_db = $this->setDatabaseConnection();
    $return = call_user_func_array('db_query', $args);
    $this->setDatabaseConnection($previous_db);
    return $return;
  }

  public function databaseResult($result) {
    $previous_db = $this->setDatabaseConnection();
    $return = db_result($result);
    $this->setDatabaseConnection($previous_db);
    return $return;
  }

  public function databaseAffectedRows() {
    $previous_db = $this->setDatabaseConnection();
    $return = db_affected_rows();
    $this->setDatabaseConnection($previous_db);
    return $return;
  }

  /**
   * Dummy function in the base class -- other classes can override to set
   * an external database.
   */
  public function setDatabaseConnection() {
  }

  public function ensureStorage() {
    if ($this->tableExists($this->schemaIdentifier())) {
      return TRUE;
    }
    else {
      return $this->addMetricToSchema();
    }
  }

  public function addMetricToSchema() {
    module_load_include('inc', 'sampler', 'sampler.api');
    $state_data = $this->sampler->buildMetricStateData();
    // Make the Sampler API aware of the schema update first.
    if (sampler_update_schema_state('update', $this->sampler->module, $this->sampler->metric, $state_data)) {
      if ($this->tableExists($this->schemaIdentifier())) {
        return TRUE;
      }
      else {
        $schema = $this->buildMetricSchema();
        $this->createTable($this->schemaIdentifier(), $schema);
        return $this->tableExists($this->schemaIdentifier());
      }
    }
    return FALSE;
  }

  public function deleteMetricFromSchema() {
    // Clear out all entries for the metric in the shared table.
    $this->databaseQuery("DELETE FROM {%s} WHERE module = '%s' AND metric = '%s'", $this->schemaIdentifier(), $this->sampler->module, $this->sampler->metric);
    // Check for remaining rows -- if any exist, assume that other metrics are
    // still using the table, and don't drop it.
    $remaining_rows = $this->databaseResult($this->databaseQuery("SELECT COUNT(timestamp) FROM {%s}", $this->schemaIdentifier()));
    if ($remaining_rows == 0) {
      $this->dropTable($this->schemaIdentifier());
    }
    return sampler_update_schema_state('drop', $this->sampler->module, $this->sampler->metric);
  }

  /**
   * Builds the column name for a value column.
   *
   * @param $key
   *   The key name of the column.
   * @param $count
   *   The ordered column number.
   */
  protected function buildValueColumnName($key, $count) {
    return "value_$count";
  }

  /**
   * Builds the Schema API description for the metric.
   *
   * @return
   *   An array that describes the schema to Schema API.
   */
  public function buildMetricSchemaDrupal() {
    return $this->buildMetricSchema();
  }

  /**
   * Builds a schema structure for a metric, in the storage plugin's form.
   *
   * @return
   *   Information describing the schema to the storage plugin.
   */
  public function buildMetricSchema() {
    $schema = array(
      'description' => t("Data for !metric metric from !module module.", array('!metric' => $this->sampler->metric, '!module' => $this->sampler->module)),
      'fields' => array(
        'module' => array(
          'type' => 'varchar',
          'length' => 255,
          'not null' => TRUE,
          'default' => '',
          'description' => t('The module providing the metric.'),
        ),
        'metric' => array(
          'type' => 'varchar',
          'length' => 50,
          'not null' => TRUE,
          'default' => '',
          'description' => t('The metric name.'),
        ),
        'object_id' => array(
          'type' => 'int',
          'unsigned' => TRUE,
          'not null' => TRUE,
          'default' => 0,
          'description' => t('Unique identifier for the object type.'),
        ),
        'timestamp' => array(
          'type' => 'int',
          'unsigned' => TRUE,
          'not null' => TRUE,
          'default' => 0,
          'description' => t('Unix timestamp representing the time the metric was calculated.'),
        ),
      ),
      'primary key' => array('module', 'metric', 'object_id', 'timestamp'),
    );
    // Build as many columns as we need depending on the data types declared.
    $count = 0;
    foreach ($this->sampler->dataType as $key => $type) {
      $count++;
      $column = $this->buildValueColumnName($key, $count);
      $schema['fields'][$column] = array(
        'type' => $type,
        'description' => t('The calculated value of the metric for the specified time period.'),
      );
    }
    return $schema;
  }

  public function getLastSampleTime() {
    $identifier = $this->schemaIdentifier();
    if ($this->tableExists($identifier)) {
      // We assume the last sample has the most recent timestamp.
      $result = $this->databaseQuery("SELECT MAX(timestamp) FROM {%s} WHERE module = '%s' AND metric = '%s'", $identifier, $this->sampler->module, $this->sampler->metric);
      if ($result && ($timestamp = $this->databaseResult($result))) {
        return intval($timestamp);
      }
    }
    return FALSE;
  }

  public function insertSamples($samples) {
    $column_placeholders = array();
    $value_placeholders = array();
    $count = 0;
    foreach ($this->sampler->dataType as $type) {
      $count++;
      $column_placeholders[] = "value_$count";
      $value_placeholders[] = db_type_placeholder($type);
    }
    $columns = "module, metric, object_id, timestamp, " . implode(', ', $column_placeholders);
    $values = "VALUES ('%s', '%s', %d, %d, " . implode(', ', $value_placeholders) . ")";

    $table_args = array($this->schemaIdentifier());

    $samples_count = 0;
    $objects = 0;
    foreach ($samples as $sample) {
      $samples_count++;
      foreach ($sample->values as $object_id => $sample_values) {
        $primary_key_args = array($this->sampler->module, $this->sampler->metric, $object_id, $sample->timestamp);
        $insert_args = array_merge($table_args, $primary_key_args, $sample_values);
        if ($this->databaseQuery("INSERT INTO {%s} ($columns) $values", $insert_args)) {
          $objects++;
        }
      }
    }
    // Inject some helpful data about the save operation into the sampler
    // object.
    $this->sampler->samplesSaved = $samples_count;
    $this->sampler->objectsSaved = $objects;

    return TRUE;
  }
}
